/*
 * Created on 25.03.2004
 */
package de.korelstar.tetris;

import java.awt.*;
import java.awt.event.*;

/** Das komplette Spielfeld als AWT-Komponente.
 * @author Kristof Hamann
 */
public class Spielfeld extends Component implements Runnable, KeyListener {
	/** Definiert die Breite des Spielfelden. */
	public static final int WIDTH = 20;
	/** Definiert die Hhe des Spielfeldes. */
	public static final int HEIGHT = 30;
	/** Definiert die Gre eines Kstchens in Pixel. */
	public static final int SIZE = 20;

	/** Das Spielfeld wird als zweidimensionales Array reprsentiert.
	 * Abgeruft wird es mit <code>spielfeld[x][y]</code>. */
	private Color[][] spielfeld = new Color[WIDTH][HEIGHT];

	/** Der aktuelle Stein, der gesteuert wird. */
	private Stein curStein;

	/** Unser Thread, der den Spielablauf regelt. */
	private Thread thread;
	
	/** Die Geschwindigkeit des Spielablaufes. In Milisekunden wird angegeben, wie lange gewartet wird bis der Stein ein Feld weiter runter geht */
	private int speed = 500;

	/** Wenn der Spieler die Leertaste drckt, wird's hier <code>true</code> und der Stein schiet nach unten */
	private boolean fast_down = false;
	
	/** Standard-Konstruktor. */
	public Spielfeld() {}

	/** Maximum-Gr&ouml;&szlig;e dieser Komponente fuer Layout-Manager. */
	public Dimension getMaximumSize() {
		return new Dimension(SIZE * WIDTH, SIZE * HEIGHT);
	}
	/** Minimum-Gr&ouml;&szlig;e dieser Komponente fuer Layout-Manager. */
	public Dimension getMinimumSize() {
		return new Dimension(SIZE * WIDTH, SIZE * HEIGHT);
	}
	/** Bevorzugte-Gr&ouml;&szlig;e dieser Komponente fuer Layout-Manager. */
	public Dimension getPreferredSize() {
		return new Dimension(SIZE * WIDTH, SIZE * HEIGHT);
	}

	/** ndert die Geschwidigkeit des Spieles.
	 * @param speed Zeit in Milisekunden, die der Thread wartet, bis der Stein ein Feld weiter nach unten wandert. */
	public void setSpeed(int speed) {
		this.speed = speed;
	}

	/**
	 * Gibt Inhalt eines Punktes im Spielfeld zurck.
	 * @param point Punkt im Spielfeld
	 * @return Status-Wert des Feldes
	 * @throws IndexOutOfBoundsException Wenn die Position auerhalb des Spielfeldes liegt
	  */
	public Color getPos(Point point) throws IndexOutOfBoundsException {
		return getPos(point.x, point.y);
	}

	/**
	 * Gibt Inhalt eines Punktes im Spielfeld zurck.
	 * @param x Position von links im Spielfeld
	 * @param y Position von unten im Spielfeld
	 * @return Status-Wert des Feldes
	 * @throws IndexOutOfBoundsException Wenn die Position auerhalb des Spielfeldes liegt
	  */
	public Color getPos(int x, int y) throws IndexOutOfBoundsException {
		if(x<0 || x>=WIDTH || y<0 || y>=HEIGHT) {
			throw new IndexOutOfBoundsException("Die Position liegt auerhalb des Spielfeldes!");
		}
		return spielfeld[x][y];
	}

	/** Speichert einen Stein im Spielfeld fest. */
	private void saveStein(Stein curStein) {
		Point[] felder = curStein.getAllPos();
		boolean verloren = false;
		for(int i=0; i<felder.length; i++) {
			if(felder[i].y>=Spielfeld.HEIGHT) {
				verloren = true;
			} else {
				spielfeld[felder[i].x][felder[i].y] = curStein.getColor();
			} 
		}
		if(verloren) {
			System.out.println("Spiel verloren");
			Toolkit.getDefaultToolkit().beep();
			// TODO den Spieler in Kentniss setzen, dass er verloren hat
			stopGame();
			return;
		}
	}
	
	/** Prft, ob eine Reihe vollgemacht wurde und lscht diese. */
	private void checkLines() {
		int i,j, x,y;
		for(i=HEIGHT-1; i>=0; i--) {
			for(j=0; j<WIDTH; j++) {
				if(spielfeld[j][i]==null) break;
				if(j==WIDTH-1) {
					//System.out.println("Volle Zeile gefunden: "+i);
					for(y=i; y<HEIGHT; y++) {
						for(x=0; x<WIDTH; x++) {
							if(y==HEIGHT-1) spielfeld[x][y] = null;
							else spielfeld[x][y] = spielfeld[x][y+1];
						}
					}
					repaintFull();
				}
			}
		}
	}

	/** Malt das komplette Spielfeld neu. */
	public void repaintFull() {
		Graphics g = getGraphics();
		paint(g);
	}

	/** Malt nur den aktuellen Stein neu. */
	public void repaintStein() {
		Graphics g = getGraphics();
		paintStein(g);
	}

	/** Wird automatisch aufgerufen, wenn das Spielfeld neu gezeichnet werden muss. */
	public void paint(Graphics g) {
		int i,j;
		g.setColor(getBackground());
		g.fillRect(0, 0, getWidth(), getHeight());
		for(i=0; i<spielfeld.length; i++) {
			for(j=0; j<spielfeld[i].length; j++) {
				if(spielfeld[i][j]!=null) {
					drawFeld(g, spielfeld[i][j], getHeight(), i, j);
				}
			}
		}
		paintStein(g);
	}
	/** Malt den aktuellen Stein neu. */
	private void paintStein(Graphics g) {
		int x,y, i;
		if(curStein != null) {
			// DoubleBuffer fr aktuellen Stein, damit es nicht flimmert
			// Er ist je ein Pixel oben, links, rechts breiter als ein Stein sein kann, um vorherige Position zu berdecken
			Image offscreenImage = createImage(SIZE*6, SIZE*5);
			if(offscreenImage==null) return; // Kann gerade nicht angezeigt werden
			Graphics offscreenGraphics = offscreenImage.getGraphics();

			//offscreenGraphics.setColor(Color.BLACK);
			//offscreenGraphics.fillRect(0, 0, SIZE*6, SIZE*5);

			// Zunchst besetzten Felder in dem Bereich des DoubleBuffer malen 
			for(x=curStein.getPos().x-2; x<curStein.getPos().x+4; x++) {
				if(x<0) continue;
				if(x>=WIDTH) continue;
				for(y=curStein.getPos().y-1; y<curStein.getPos().y+4; y++) {
					if(y<0) continue;
					if(y>=HEIGHT) continue;
					drawFeld(offscreenGraphics, spielfeld[x][y], SIZE*5, x - curStein.getPos().x+2, y - curStein.getPos().y+1);
				}//
			}

			// Dann den Stein an seiner Position in der DoubleBuffer malen
			Point[] stein = curStein.getAllPos();
			for(i=0; i<stein.length; i++) {
				drawFeld(offscreenGraphics, curStein.getColor(), SIZE*5, stein[i].x-curStein.getPos().x+2, stein[i].y-curStein.getPos().y+1);
			}

			// DoubleBuffer in das Spielfeld ausgeben
			g.drawImage(offscreenImage, (curStein.getPos().x-2)*SIZE, getHeight() - (curStein.getPos().y+4)*SIZE, this);
		}
	}

	/** Zeichnet ein Feld in Form eines Quadrates an die angegebene Position. */
	private void drawFeld(Graphics g, Color farbe, int height, Point punkt) {
		drawFeld(g, farbe, height, punkt.x, punkt.y);
	}
	private void drawFeld(Graphics g, Color farbe, int height, int x, int y) {
		if(farbe==null) {
			//g.setColor(getBackground());
			//g.fillRect(x*SIZE, getHeight() - ((y+1)*SIZE), SIZE, SIZE);
		} else {
			g.setColor(Color.BLACK);
			g.drawRect(x*SIZE+1, height - ((y+1)*SIZE)+1, SIZE-2, SIZE-2);
			g.setColor(farbe);
			g.fillRect(x*SIZE+1, height - ((y+1)*SIZE)+1, SIZE-2, SIZE-2);
		}
	}


	/** Startet ein neues Spiel auf einem leeren Brett. */
	public void startGame() {
		int x,y;
		for(x=0; x<Spielfeld.WIDTH; x++) {
			for(y=0; y<Spielfeld.HEIGHT; y++) {
				spielfeld[x][y] = null;
			}
		}
		curStein = null;
		thread = new Thread(this);
		thread.start();
		repaintFull();
	}

	/** Stoppt das aktuelle Spiel. z.B. bei GameOver. */
	public void stopGame() {
		curStein = null;
		thread = null;
	}

	/** Thread, der die Spielsteine nach unten bewegt */
	public void run() {
		Thread thisThread = Thread.currentThread();
		boolean amBoden = false;
		while(thread == thisThread) {
			if(curStein == null) {
				curStein = new Stein(this);
			} else {
				try {
					curStein.moveDown();
					amBoden = false;
				} catch(IndexOutOfBoundsException e) {
					if(amBoden) {
						//System.out.println("Stein ist endgltig auf dem Boden angekommen. Schmeie neuen raus...");
						saveStein(curStein);
						curStein = null;
						checkLines();
						curStein = new Stein(this, Stein.STEIN_RAND);
						fast_down = false;
					} else {
						amBoden = true;
						//System.out.println("Stein ist auf dem Boden angekommen. Warte noch eine Runde, ob sich was tut...");
						if(!fast_down) {
							try {
								Thread.sleep(100);
							} catch(InterruptedException ie) {}
						}
					}
				}
			}
			if(!fast_down) {
				try {
					Thread.sleep(speed);
				} catch(InterruptedException e) {}
			}
		}
	}

	// KeyListener
	public void keyTyped(KeyEvent e) { }
	public void keyPressed(KeyEvent e) { }
	public void keyReleased(KeyEvent e)
	{
		int key = e.getKeyCode();
		try {
			if(key == KeyEvent.VK_UP) curStein.turn();
			if(key == KeyEvent.VK_DOWN) curStein.moveDown();
			if(key == KeyEvent.VK_LEFT) curStein.moveLeft();
			if(key == KeyEvent.VK_RIGHT) curStein.moveRight();
			if(key == KeyEvent.VK_SPACE) {
				fast_down = true;
				//thread.notifyAll();
			}
		} catch(IndexOutOfBoundsException ex) {
			//Toolkit.getDefaultToolkit().beep();
		}
	}
}