/*
 * Created on 25.03.2004
 */
package de.korelstar.tetris;

import java.awt.Point;
import java.awt.Color;

/** Reprsentiert einen einzelnen Tetris-Stein.
 * @author Kristof Hamann
 */
public class Stein {

	/** Wert zum Erstellen eines Steines mit zuflligem Typ. */
	public static final int STEIN_RAND = -1;

	/** Enth&auml;lt die relative Positionen der Felder der verschiedenen Steine */
	private static final Point[][] STEINE = {
		{ new Point(-1,  1), new Point( 1,  1), new Point( 2,  1), new Point( 1, -1) }, // "N" Spiegel
		{ new Point(-2,  1), new Point(-1,  1), new Point( 1,  1), new Point( 1, -1) }, // L
		{ new Point( 2,  1), new Point( 1,  1), new Point(-1,  1), new Point(-1, -1) }, // L Spiegel
		{ new Point(-1,  1), new Point( 1,  1), new Point( 1, -1), new Point(-1, -1) }, // Quadrat
		{ new Point(-1, -1), new Point( 1, -1), new Point( 1,  1), new Point( 2,  1) }, // "N"
		{ new Point(-1,  1), new Point( 1,  1), new Point( 1, -1), new Point( 2, -1) }, // "N" Spiegel
		{ new Point(-2,  1), new Point(-1,  1), new Point( 1,  1), new Point( 2,  1) }, // Tetris
	};
	
	/** Enth&auml;lt die Farben der verschiedenen Steine. */
	private static final Color[] FARBEN = {
		Color.RED,
		Color.GREEN,
		Color.BLUE,
		Color.YELLOW,
		Color.ORANGE,
		Color.DARK_GRAY,
		Color.CYAN
	};

	/** Speichert die relative Position der verschiedenen Felder des Steines. */
	private Point[] felder;
	/** Speichert die absolute Position des Steines. */
	private Point position;
	/** Spielfeld, auf welchem sich der Stein befindet. */
	private Spielfeld spielfeld;
	/** Die Farbe des Steines. */
	private Color farbe;


	/** Erstellt neuen Stein. Der Typ wird hierbei zufllig gewhlt.
	 * @param spielfeld Spielfeld, auf welchem der Stein erscheinen soll 
	 */
	Stein(Spielfeld spielfeld) {
		this(spielfeld, STEIN_RAND);
	}
	/** Erstellt neuen Stein.
	 * <p>
	 * Der Parameter <code>typ</code> kann hierbei einen der folgenden Werte
	 * annehmen: 
     * <ul>
     * <li><code>STEIN_RAND</code>, 
     * <li>einen Index-Wert des konstanten Arrays <code>STEIN</code> bzw. <code>FARBEN</code>. 
     * </ul>
	 *
	 * @param spielfeld Spielfeld, auf welchem der Stein erscheinen soll 
	 * @param typ Eine Konstante, die den Typ des Steines angbit 
	 */
	Stein(Spielfeld spielfeld, int typ) {
		if(typ == STEIN_RAND) {
			typ = (int)(Math.random() * STEINE.length);
		}
		this.position = new Point(Spielfeld.WIDTH/2, Spielfeld.HEIGHT-1);
		this.felder = STEINE[typ];
		this.farbe = FARBEN[typ];
		this.spielfeld = spielfeld;
	}


	/** Gibt die Farbe des Steines zurck.
	 * @return Farb-Objekt des Steines
	 */
	public Color getColor() {
		return farbe;
	}

	/** Gibt die aktuelle Position des Steines zurck.
	 * @return Position des Steines
	 */
	public Point getPos() {
		return position;
	}

	/** Gibt die aktuelle Position des Steines zurck.
	 * @return Position des Steines
	 */
	public Point[] getFelder() {
		return felder;
	}

	/** Bewegt den Stein ein Feld nach unten.
	 * @throws IndexOutOfBoundsException Wenn am Rand oder an einem anderen Stein angekommen 
	 */
	public void moveDown() throws IndexOutOfBoundsException {
		tryMove(felder, new Point(position.x, position.y-1));
	}

	/** Bewegt den Stein ein Feld nach rechts. */
	public void moveRight() {
		tryMove(felder, new Point(position.x+1, position.y));
	}

	/** Bewegt den Stein ein Feld nach links. */
	public void moveLeft() {
		tryMove(felder, new Point(position.x-1, position.y));
	}

	/** Dreht den Stein im Uhrzeigersinn. */
	public void turn() {
		Point[] felderNeu = new Point[felder.length];
		for(int i=0; i<felder.length; i++) {
			felderNeu[i] = new Point(felder[i].y, -1 * felder[i].x);
		}
		tryMove(felderNeu, position);
	}

	/**
	 * Gibt die absolute Position dieses Steines im Spielfeld zurck.
	 * @return Array aus Punkten mit der absoluten Position dieses Steines
	 */
	public Point[] getAllPos() {
		Point[] felderAbs = new Point[felder.length];
		for(int i=0; i<felder.length; i++) {
			felderAbs[i] = getPos(felder[i], position);
		}
		return felderAbs;
	}

	/**
	 * Gibt die absolute Position im Spielfeld zu einem Feld des Steines zurck.
	 * @param feld Relative Position des Feldes
	 * @param position Absolute Position des Steines im Spielfeld
	 * @return Punkt mit der absoluten Position dieses Feldes
	 */
	static private Point getPos(Point feld, Point position) {
		Point punkt = new Point();
		if(feld.x>0) punkt.x = feld.x + position.x;
		if(feld.x<0) punkt.x = feld.x + position.x + 1;
		if(feld.y>0) punkt.y = feld.y + position.y;
		if(feld.y<0) punkt.y = feld.y + position.y + 1;
		return punkt;
	}

	/** Fhrt eine Bewegung des Steines aus.
	 * @param felder Array mit relativen Positionen der Felder des Steines
	 * @param position Punkt, der die absolute Position des Steines im Spielfeld angibt
	 * @throws IndexOutOfBoundsException wenn der Zug des Steines nicht machbar ist
	 */
	private synchronized void tryMove(Point[] felder, Point position) throws IndexOutOfBoundsException {
		Point absPos;
		for(int i=0; i<felder.length; i++) {
			absPos = getPos(felder[i], position);
			// Prft, ob unten angekommen
			if(absPos.y < 0) throw new IndexOutOfBoundsException("Stein ist am unteren Rand angekommen.");
			// Prft, ob links angekommen
			if(absPos.x < 0) throw new IndexOutOfBoundsException("Stein ist am linken Rand angekommen.");
			// Prft, ob rechts angekommen
			if(absPos.x >= Spielfeld.WIDTH) throw new IndexOutOfBoundsException("Stein ist am rechten Rand angekommen.");
			// Prft, ob noch ber dem Spielfeld, dann nmlich beim nchsten Stein weitermachen
			if(absPos.y >= Spielfeld.HEIGHT) continue;
			// Prft, ob auf dem Feld ein Stein liegt
			if(spielfeld.getPos(absPos.x, absPos.y)!=null) throw new IndexOutOfBoundsException("Stein ist auf einen anderen Stein getroffen.");
		}
		this.felder = felder;
		this.position = position;
		spielfeld.repaintStein();
	}
}